#include "NekoDriver.h"
extern "C" {
#include "ANSI/65c02.h"
}
#include "CC800IOName.h"
#include "log.h"


// Storage
PNekoDriver theNekoDriver;

TNekoDriver::TNekoDriver()
    : fNorBuffer(NULL)
    , fBROMBuffer(NULL)
    , fLCDBufferChanged(NULL)
{
    // Do initialization that must be repeated for a restart
    restart = 0;

    fBROMBuffer = (char*)malloc(512 * 0x8000); // 32K * 256 * 2
    for (int i = 0; i < 256; i++) {
        volume0array[i] = (unsigned char*)fBROMBuffer + i * 0x8000;
        volume1array[i] = volume0array[i] + 256 * 0x8000;
    }

    fNorBuffer = (char*)malloc(16 * 0x8000); // 32K * 16
    for (int i = 0; i < 16; i++) {
        norbankheader[i] = (unsigned char*)fNorBuffer + i * 0x8000;
    }

    //         DebugInitialize();
    MemInitialize();
    //         TermInitialize();
    //         FrameCreateWindow();

    // enter the main message loop
    //         EnterMessageLoop();

//     fEmulatorThread->start();
}


TNekoDriver::~TNekoDriver()
{
    if (fNorBuffer) {
        free(fNorBuffer);
        fNorBuffer = NULL;
    }
    if (fBROMBuffer) {
        free(fBROMBuffer);
        fBROMBuffer = NULL;
    }
}


void TNekoDriver::SetOnLCDBufferChanged(OnLCDBufferChanged lcdBufferChanged)
{
    fLCDBufferChanged = lcdBufferChanged;

}

void TNekoDriver::OnStepFinished( unsigned short pc )
{
}

bool TNekoDriver::StartEmulation()
{
    fEmulatorThread.SetOnLCDBufferChanged(fLCDBufferChanged);   
    fEmulatorThread.start();
    return true;
}

bool TNekoDriver::StopEmulation()
{
    fEmulatorThread.interrupt();
    return true;
}

bool TNekoDriver::RunDemoBin( const QString& filename )
{
    LoadBROM(filename + "/obj.bin");
    LoadFullNorFlash(filename + "/cc800.fls");
    StopEmulation();
    StartEmulation();   
    return true;
}

bool TNekoDriver::PauseEmulation()
{
    return false;
}

bool TNekoDriver::ResumeEmulation()
{
    return false;
}



EmulatorThread::EmulatorThread()
    : fLCDBuffer(malloc(160*80/8))
    , lastTicket(0)
    , totalcycle(0)
    , checked(false)
    , batchlimiter(0)
    , batchcount(UINT_MAX)
    , sleepgap(10)
    , sleepcount(0)
    , fLCDBufferChanged(NULL)
{

}

EmulatorThread::~EmulatorThread()
{
    free(fLCDBuffer);
}

void EmulatorThread::SetOnLCDBufferChanged(OnLCDBufferChanged lcdBufferChanged)
{
    fLCDBufferChanged = lcdBufferChanged;
}


extern WORD LogDisassembly (WORD offset, LPTSTR text);
extern void AppendLog(const char* text);

extern bool timer0started;
extern bool timer1started;
extern unsigned short gThreadFlags;

//DWORD lastTicket = 0;
//unsigned long long totalcycle = 0;
//const unsigned spdc1016freq = 3686400;
//bool checked = false;
//unsigned batchlimiter = 0;
//long batchcount = LONG_MAX;
//double sleepgap = 10;
//long sleepcount = 0;

// WQXSIM
extern bool timer0waveoutstart;
extern int prevtimer0value;
int gDeadlockCounter = 0;
extern bool lcdoffshift0flag;
bool matrixupdated = false;
long nmicount = 0;

void Turnoff2HzNMIMaskAddIRQFlag();
void CheckTimebaseAndEnableIRQnEXIE1();
void EnableWatchDogFlag();
void CheckLCDOffShift0AndEnableWatchDog();

const unsigned long spdc1016freq = 3686400;
bool shouldNMI = false;
bool shouldIRQ = false;


const int FrameRate = 30;
const unsigned long CyclesPerFrame = spdc1016freq / FrameRate;
const unsigned long CyclesPerNMI = spdc1016freq / 2;
const unsigned long CyclesPer10Ms = spdc1016freq / 100; 

unsigned int frameCounter = 0;
unsigned int nmiCounter = 0;
unsigned int clockCounter = 0;
unsigned int totalInsts = 0;

void EmulatorThread::main()
{
    unsigned long frameCycles = CyclesPerFrame * (frameCounter + 1);
    unsigned long nmiCycles = CyclesPerNMI * (nmiCounter + 1);
    unsigned long clockCycles = CyclesPer10Ms * (clockCounter + 1);  
    while (totcycles < frameCycles && !isInterrupted()) {
        //LOGD("cpu cycles = %d\n", totcycles);
        if (shouldNMI) {
            nmi = 0;
            shouldNMI = false;
        } else if(shouldIRQ && !(regs.ps & 0x04)) {//ERROR XXX 555990 IRQ
            irq = 0;
            shouldIRQ = false;
        }
             
        DWORD CpuTicks = CpuExecute();
        totcycles += CpuTicks;
        
        if (totcycles >= nmiCycles) {
            nmiCounter++;
            nmiCycles += CyclesPerNMI;
            shouldNMI = true;
        }
        if (totcycles >= clockCycles) {
            clockCounter++;
            clockCycles += CyclesPer10Ms;
            CheckTimebaseAndEnableIRQnEXIE1();
        }
        //if (this.timer0started) {
        //    this.shouldIrq = true;
        //}
        totalInsts++;
    }
    
}
/*
#if TARGET_OS_ANDROID
#include <time.h>
#include <errno.h>

int Sleep(unsigned long usec)
{
  struct timespec ts;

  ts.tv_sec  = usec/1000000UL;

#ifdef __arm__
  ts.tv_nsec = (usec - ts.tv_sec*1000000UL)*1000;
#else
  ts.tv_nsec = (usec % 1000000UL) * 1000UL;
#endif

  for (;;)
  {
    if ( nanosleep( &ts, &ts ) == 0 )
        return 0;

    // We try again if the nanosleep failure is EINTR.
    // The other possible failures are EINVAL (which we should pass through),
    // and ENOSYS, which doesn't happen.
    if ( errno != EINTR )
        return -1;
  }
}
#endif
*/
#include <unistd.h>

void EmulatorThread::run()
{
    LOGD("CpuInitialize");
    CpuInitialize();
    int timerPerFrame = 1000/FrameRate*1000;
	LOGD("CpuExecute %d", isInterrupted());
    while (!isInterrupted()) {
        long time = clock();
        main();
		if (memcmp(fLCDBuffer, &fixedram0000[0x9C0], 160*80/8) != 0) {
            memcpy(fLCDBuffer, &fixedram0000[0x9C0], 160*80/8);
            if (fLCDBufferChanged != NULL) fLCDBufferChanged((const char*)fLCDBuffer);
		}
        frameCounter++;
		time = clock() - time;
        if (time < timerPerFrame) {
            usleep(timerPerFrame - time);
        }
    }
 }

void CheckLCDOffShift0AndEnableWatchDog()
{
    //// may check hotkey press
    //if ( gLcdoffShift0Flag )
    //{
    //    if ( keypadmatrix1[6] || keypadmatrix1[7] )
    //    {
    //        // Line6 Dict Card
    //        // Line7 on/off
    //        EnableWatchDogFlag();               // 0x80
    //        gLcdoffShift0Flag = 0;
    //    }
    //}
    //else
    //{
    //    // Set lcdoffshift0flag only when keypadmatrix3 == 0xFB
    //    if ( keypadmatrix1[7] == 0xFBu )
    //        gLcdoffShift0Flag = 1;
    //}
    matrixupdated = true;
    if (lcdoffshift0flag) {
        // we don't have invert bit for row6,7
        //bool row67down = false;
        for (int y = 0; y < 2; y++) {
            for (int x = 0; x < 8; x++) {
                if (keypadmatrix[y][x] == 1) {
                    //row6,7down = true;
                    EnableWatchDogFlag();
                    lcdoffshift0flag = false;
                    return;
                }
            }
        }
    } else {
        if (keypadmatrix[0][2] == 1) {
            lcdoffshift0flag = true; // this flag is used for UI?
        }
    }
}

void CheckTimebaseAndEnableIRQnEXIE1()
{
    //if ( gFixedRAM0_04_general_ctrl & 0xF )
    //{
    //    // TimeBase Clock Select bit0~3
    //    LOBYTE(gThreadFlags) = gThreadFlags | 0x10; // add 0x10 to gThreadFlags
    //    gFixedRAM0[(unsigned __int8)io01_int_ctrl] |= 8u;// EXTERNAL INTERRUPT SELECT1
    //}
    if (fixedram0000[io04_general_ctrl] & 0x0F) {
        gThreadFlags |= 0x10; // Add IRQ flag
        //irq = 0; // TODO: move to NMI check
        //FIXME
        shouldIRQ = true;
        //printf("%d \n", totalInsts);
        fixedram0000[io01_int_enable] |= 0x8; // EXTERNAL INTERRUPT SELECT1
    }
}

void Turnoff2HzNMIMaskAddIRQFlag()
{
    if ( fixedram0000[io04_general_ctrl] & 0xF )
    {
        gThreadFlags |= 0x10u; // Add 0x10 Flag to gThreadFlag
        //irq = 0; // TODO: move to 
        fixedram0000[io01_int_enable] |= 0x10u; // 2Hz NMI Mask off
    }
}

void EnableWatchDogFlag()
{
    gThreadFlags |= 0x80;
}
