#include "NekoDriver.h"
#include "log.h"
#include <jni.h>
#include <stdlib.h>

#define EMULATOR_CLS "com/emu/cc800/Emulator"
#define BUFFER_SIZE 160*80/8
#define C_DECL extern "C"
#define FUN_DECL(name) Java_com_emu_cc800_Emulator_##name


static JavaVM *g_pVM = NULL;
static jmethodID g_methodID = NULL;
static jfieldID g_fieldID = NULL;
static jbyteArray g_array = NULL;
static jobject g_objEmulator = NULL;
static jclass g_clsEmulator = NULL;

static TNekoDriver theDriver;

jobject newGlobalRef(JNIEnv *env, jobject ref) 
{
    jobject obj = env->NewGlobalRef(ref);
    env->DeleteLocalRef((jobject)ref);
    return obj;
}

C_DECL void onLCDBufferChanged(const char* buffer)
{
   if (g_pVM == NULL) {
      LOGE("g_pVM is invalid.");
      return;
   }
   JNIEnv *pEnv = NULL;
   jint result = g_pVM->AttachCurrentThread(&pEnv, NULL);
   if (result != JNI_OK) {
      LOGE("g_pVM attach current thread failed.");
	  return;
   }
   jbyte* data = pEnv->GetByteArrayElements(g_array, NULL); 
   memcpy(data, buffer, BUFFER_SIZE);
   pEnv->CallVoidMethod(g_objEmulator, g_methodID);
   pEnv->ReleaseByteArrayElements(g_array, data, 0);
   g_pVM->DetachCurrentThread();   
}

C_DECL jint FUN_DECL(start)(JNIEnv *env, jobject obj, jstring sDirName)
{
   const char* szDirName = env->GetStringUTFChars(sDirName, NULL);
   LOGD("start emulator: %s", szDirName);
   theDriver.SetOnLCDBufferChanged(onLCDBufferChanged);
   theDriver.RunDemoBin(szDirName);
   env->ReleaseStringUTFChars(sDirName, szDirName);
   return 0;
}

C_DECL jint FUN_DECL(reset)(JNIEnv *env, jobject obj)
{
   LOGD("reset emulator");
   theDriver.ResumeEmulation();
   return 0;
}

C_DECL jboolean FUN_DECL(isStarted)(JNIEnv *env, jobject obj)
{
   return theDriver.IsStarted();
}

C_DECL jint FUN_DECL(stop)(JNIEnv *env, jobject obj)
{
   LOGD("stop emulator");
   theDriver.StopEmulation();
   return 0;
}


C_DECL jint FUN_DECL(attach)(JNIEnv *env, jobject obj, jobject objEmu)
{
    LOGD("attach emulator: 0x%04x", objEmu);
    g_objEmulator = env->NewGlobalRef(objEmu);
    
    jclass cls = env->FindClass(EMULATOR_CLS);
    g_clsEmulator = (jclass)newGlobalRef(env, cls);
    LOGD("class = %d", g_clsEmulator);
    
    g_methodID= env->GetMethodID(g_clsEmulator, "drawScreen", "()V");
    LOGD("methoID = %d", g_methodID);
    
    
    g_fieldID  = env->GetFieldID(g_clsEmulator, "mData", "[B");
    LOGD("fieldID = %d", g_fieldID);
    
    jbyteArray array = (jbyteArray)env->GetObjectField(g_objEmulator, g_fieldID);
    g_array = (jbyteArray)newGlobalRef(env, array);
    LOGD("obj = %d", g_array);    
    
    return 0;
}

C_DECL void FUN_DECL(detach)(JNIEnv *env, jobject obj) 
{
    if (g_objEmulator != NULL) {
        env->DeleteGlobalRef(g_objEmulator);
        g_objEmulator = NULL;
    }
    if (g_array != NULL) {
        env->DeleteGlobalRef(g_array);
        g_array = NULL;
    }
}


C_DECL jint FUN_DECL(keyPress)(JNIEnv *env, jobject obj, jint index)
{
   LOGD("keyPress %d", index);
   *((char*)keypadmatrix+index) = 1;
   return 0;
}

C_DECL jint FUN_DECL(keyRelease)(JNIEnv *env, jobject obj, jint index)
{
   LOGD("keyRelease %d", index);
   *((char*)keypadmatrix+index) = 0;
   return 0;
}

JNIEXPORT jint JNI_OnLoad(JavaVM* vm, void* reserved)
{ 
   g_pVM = vm;
   JNIEnv* env = NULL;
   if (vm->GetEnv((void**)&env, JNI_VERSION_1_2) != JNI_OK) {
      return -1;
   }
   return JNI_VERSION_1_2;
}



