package com.emu.cc800;

import java.nio.IntBuffer;

import android.app.Activity;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Matrix;
import android.os.Environment;
import android.view.SurfaceHolder;
import android.view.SurfaceHolder.Callback;

import com.emu.cc800.KeyBoard.OnKeyListener;

public class Emulator implements Callback, OnKeyListener {
	private static Emulator mInstance;
	private static final int BUFFER_SIZE = 160*80/8;
	private static final String ROM_PATH = "/cc800/";
	static {
		System.loadLibrary("CC800Android");
	}	
	private SurfaceHolder mSurfaceHolder;
	private byte[] mData;
	private int[] mScreenData;
	private Matrix mMatrix;
	private Bitmap mBitmap;
	
	private native int start(String sDirName);
	private native boolean isStarted();
	public native int reset();
	public native int stop();
	public native int keyPress(int index);
	public native int keyRelease(int index);
	private native int attach(Emulator emu);
	private native void detach();
	
	public static Emulator getInstance() {
		if (mInstance == null) {
			mInstance = new Emulator();
		}
		return mInstance;
	}
	
	private Emulator() {
		mData = new byte[BUFFER_SIZE];
		mBitmap = Bitmap.createBitmap(160, 80, Bitmap.Config.ARGB_8888);
		mScreenData = new int[BUFFER_SIZE*8];
		mMatrix = new Matrix();
		mMatrix.setScale(3, 3);
		attach(this);
	}
	
	@Override
	protected void finalize() throws Throwable {
		detach();
		super.finalize();
	}
	
	public int start() {
		return start(Environment.getExternalStorageDirectory().getPath() + ROM_PATH);
	}
	
	public void setSurface(SurfaceHolder holder) {
		mSurfaceHolder = holder;
		mSurfaceHolder.addCallback(this);
	}
	
	public void clearScreen(int color) {
		if (mSurfaceHolder != null) {
			Canvas c = null;
			try {
				c = mSurfaceHolder.lockCanvas();
				c.drawColor(color);
			}catch (Exception e) {
			}finally {
				if (c != null)
				mSurfaceHolder.unlockCanvasAndPost(c);
			}
		}
	}
	
	public void drawScreen() {
		if (mSurfaceHolder != null) {
			Canvas c = mSurfaceHolder.lockCanvas();
			if (c == null) return;
			//Paint paint = new Paint();
			//c.drawColor(0xFF00FF00);
			//to RGB888
			for (int i = 0; i < 80; i++) {
				for (int j = 0; j < 20; j++) {
					byte p = mData[i*20+j];					
					for (int z = 8; z > 0; z--) {
						int v = p & (1 << (8 - z));
						/*
						if (v == 0) paint.setColor(0xFF00FF00);
						else paint.setColor(0xFF000000);
						int left = (j*8 + z - 1)*3;
						int top = i*3;
						if (z == 1) {
							if (j>0) c.drawRect(left, top, left+3, top+3, paint);
						}
						else c.drawRect(left, top, left+3, top+3, paint);
						*/					
						mScreenData[i*160+j*8+z-1] = v == 0 ? 0xFF00FF00 : 0xFF000000;
						if (z == 1 && j == 0) mScreenData[i*160+j*8+z-1] = 0xFF00FF00;
					} 
				}
			}
			mBitmap.copyPixelsFromBuffer(IntBuffer.wrap(mScreenData));
			c.drawBitmap(mBitmap, mMatrix, null);
			mSurfaceHolder.unlockCanvasAndPost(c);
		}
	}
	
	@Override
	public void surfaceChanged(SurfaceHolder holder, int format, int width,
			int height) {
	}
	
	@Override
	public void surfaceCreated(SurfaceHolder holder) {
		clearScreen(0xFF00FF00);
	}
	
	@Override
	public void surfaceDestroyed(SurfaceHolder holder) {
	}
	
	@Override
	public void onKeyDown(int index) {
		if (index == 2) {
			if (isStarted()) {
				stop();
				clearScreen(Color.GREEN);
			}
			else start();
			return;
		}
		keyPress(index);
	}
	
	@Override
	public void onKeyUp(int index) {
		keyRelease(index);
	}
}
