package com.emu.cc800;

import android.content.Context;
import android.graphics.Color;
import android.util.AttributeSet;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import android.widget.Button;
import android.widget.LinearLayout;

public class KeyBoard extends LinearLayout implements android.view.View.OnTouchListener {
	public interface OnKeyListener {
		void onKeyDown(int index);
		void onKeyUp(int index);
	}
	
	private OnKeyListener mKeyListener;
	
	private class KeyItem {
	   public int keyCode;
	   public String keyValue;
	   public int index;
	   public int mapping;
	   
	   public KeyItem(int keyCode, String keyValue) {
		   this.keyCode = keyCode;
		   this.keyValue = keyValue;
		   this.index = -1;
		   this.mapping = -1;
	   }		
	   
	   public KeyItem(int keyCode, String keyValue, int index) {
		   this.keyCode = keyCode;
		   this.keyValue = keyValue;
		   this.index = index;
		   this.mapping = -1;
	   }	   
	   
	   public KeyItem(int keyCode, String keyValue, int index, int mapping) {
		   this.keyCode = keyCode;
		   this.keyValue = keyValue;
		   this.index = index;
		   this.mapping = mapping;
	   }
	};
	
	private KeyItem[] mItems = {
		   new KeyItem(0, "Dict", 8),
		   new KeyItem(-1, null), 
		   new KeyItem(2, "Calc", 10),
		   new KeyItem(3, "Rem", 11),
		   new KeyItem(4, "Data", 12),
		   new KeyItem(5, "Time", 13),
		   new KeyItem(6, "Net", 14),
		   new KeyItem(-1, null), 
		   new KeyItem(-1, null),
		   new KeyItem(-1, null),
		    
		   new KeyItem(-1, null),
		   new KeyItem(-1, null),
		   new KeyItem(12, "F1", 58),
		   new KeyItem(13, "F2", 59),
		   new KeyItem(14, "F3", 60),
		   new KeyItem(15, "F4", 61),
		   new KeyItem(1, "F5", 9),
		   new KeyItem(-1, null),
		   new KeyItem(18, "On", 2), 
		   new KeyItem(-1, null),
		    
		   new KeyItem(20, "Q", 40, 45),
		   new KeyItem(21, "W", 41, 51),
		   new KeyItem(22, "E", 42, 33),
		   new KeyItem(23, "R", 43, 46),
		   new KeyItem(24, "T", 44, 48),
		   new KeyItem(25, "Y", 45, 53),
		   new KeyItem(26, "U", 46, 49),
		   new KeyItem(27, "I", 47, 37),
		   new KeyItem(28, "O", 48, 43),
		   new KeyItem(29, "P", 52, 44),
		    
		   new KeyItem(30, "A", 32, 29),
		   new KeyItem(31, "S", 33, 47),
		   new KeyItem(32, "D", 34, 32),
		   new KeyItem(33, "F", 35, 34),
		   new KeyItem(34, "G", 36, 35),
		   new KeyItem(35, "H", 37, 36),
		   new KeyItem(36, "J", 38, 38),
		   new KeyItem(37, "K", 39, 39), 
		   new KeyItem(38, "L", 49, 40),
		   new KeyItem(39, "ok", 53, 66),
		    
		   new KeyItem(40, "Z", 24, 54),
		   new KeyItem(41, "X", 25, 52),
		   new KeyItem(42, "C", 26, 31),
		   new KeyItem(43, "V", 27, 50),
		   new KeyItem(44, "B", 28, 30),
		   new KeyItem(45, "N", 29, 42),
		   new KeyItem(46, "M", 30, 41),
		   new KeyItem(47, "Pup", 31),
		   new KeyItem(48, "↑", 50, 19),
		   new KeyItem(49, "Pdn", 51),
		    
		   new KeyItem(50, "?", 16),
		   new KeyItem(51, "Shift", 17),
		   new KeyItem(52, "Cap", 18, 84),
		   new KeyItem(53, "Esc", 19, 4),
		   new KeyItem(54, "0", 20),
		   new KeyItem(55, ".", 21, 56),
		   new KeyItem(56, "=", 22, 62),
		   new KeyItem(57, "←", 23, 21), 
		   new KeyItem(58, "↓", 51, 20),
		   new KeyItem(59, "→", 55, 22)    		
	};

	public void setOnKeyListener(OnKeyListener listener) {
		mKeyListener = listener;
	}
	
	public KeyBoard(Context context, AttributeSet attrs, int defStyle) {
		super(context, attrs);
		initKeyBoard();
	}
	
	public KeyBoard(Context context, AttributeSet attrs) {
		super(context, attrs);
		initKeyBoard();
	}	
	
	public KeyBoard(Context context) {
		super(context);
		initKeyBoard();
	}		

	private void initKeyBoard() {
		this.setOrientation(LinearLayout.VERTICAL);
		LayoutParams params = new LayoutParams(0,
				LayoutParams.WRAP_CONTENT);
		params.weight = 1;
		params.leftMargin = 0;
		params.rightMargin = 0;
		for (int j = 0; j < 6; j++) {
			LinearLayout linear = new LinearLayout(getContext());
			for (int i = 0; i < 10; i++) {
				int index = i+j*10;
				if (mItems[index].keyCode == -1) continue;
				Button item = new Button(getContext());
				item.setOnTouchListener(this);
				item.setFocusable(false);
				item.setPadding(0, 0, 0, 0);
				//item.setLayoutParams(params);
				//item.setBackgroundResource(R.drawable.btn_selector);
				//item.setTextSize(12);
				item.setTag(mItems[index]);
				item.setText(mItems[index].keyValue);
				linear.addView(item, params);
			}
			this.addView(linear);
		}
	}
	
	public boolean onKeyEvent(int keyCode, int action) {
		for (KeyItem item : mItems) {
			if (item.mapping == keyCode) {
				if (action == KeyEvent.ACTION_DOWN)
					mKeyListener.onKeyDown(item.index);
				else
					mKeyListener.onKeyUp(item.index);
				return true;
			}
		}
		return false;
	}

	@Override
	public boolean onTouch(View v, MotionEvent event) {
		KeyItem item = (KeyItem) v.getTag();
		if (mKeyListener == null) return false;
		int action = event.getAction();
		if (action == MotionEvent.ACTION_DOWN)
			mKeyListener.onKeyDown(item.index);
		else if (action == MotionEvent.ACTION_UP || action == MotionEvent.ACTION_CANCEL)
			mKeyListener.onKeyUp(item.index);
		return false;
	}
}
