//
//  CCKeyboard.m
//  Sim800IOS

#import "CCKeyboard.h"
#include "KeyBoard.h"

#define KEY_WIDTH 32
#define KEY_HEIGHT 30

@implementation CCKeyboard

@synthesize keyboardDelegate;

- (id)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        // Initialization code
    }
    return self;
}


// Only override drawRect: if you perform custom drawing.
// An empty implementation adversely affects performance during animation.
- (void)drawRect:(CGRect)rect
{
    // Drawing code
    CGContextRef c = UIGraphicsGetCurrentContext();
    CGContextBeginPath(c);
    CGContextSetTextMatrix(c, CGAffineTransformMakeScale(1.0, -1.0));
    const char* fontName = [[UIFont systemFontOfSize:[UIFont labelFontSize]].fontName UTF8String];
    for (int i=0; i<10; i++) {
        for (int j=0; j<6; j++) {
            KeyItem item = keyMappingTable[i+j*10];
            if (item.keyCode == INVALID_KEY) {
                continue;
            }
            CGContextSetRGBStrokeColor(c, 0, 1, 0, 1);
            CGContextSetRGBFillColor(c, 0, 1, 0, 1);
            CGContextAddRect(c, CGRectMake(KEY_WIDTH*i+1, KEY_HEIGHT*j+2, 30, 26));
            CGContextSetRGBFillColor(c, 0, 0, 0, 1);
            CGContextSelectFont(c, fontName, 12, kCGEncodingMacRoman);
            CGContextSetTextDrawingMode(c, kCGTextFill);
            CGContextShowTextAtPoint(c, KEY_WIDTH*i+2, KEY_HEIGHT*j+12, item.keyValue, strlen(item.keyValue));
        }
    }
    CGContextDrawPath(c, kCGPathStroke);
    CGContextClosePath(c);
}

-(void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
    for (UITouch *touch in touches) {
        CGPoint point = [touch locationInView:self];
        int x = point.x/32;
        int y = point.y/30;
        CGRect rect = CGRectMake(KEY_WIDTH*x+1, KEY_HEIGHT*y+2, 30, 26);
        if (CGRectContainsPoint(rect, point)) {
            NSLog(@"x,y = %d, %d \n", x, y);
            KeyItem item = keyMappingTable[x+y*10];
            if (item.keyCode != INVALID_KEY) {
                NSLog(@"KeyCode: %s\n", item.keyValue);
                keypadmatrix[item.index/8][item.index%8] = 1;
            }
            if (item.index == 2) {
                if ([self.keyboardDelegate conformsToProtocol:@protocol(CCKeyboardDelegate)]) {
                    [self.keyboardDelegate performSelector:@selector(onPower)];
                }
            }
        }
    }
}

-(void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event
{
}

-(void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
    for (UITouch *touch in touches) {
        CGPoint point = [touch locationInView:self];
        int x = point.x/32;
        int y = point.y/30;
        CGRect rect = CGRectMake(KEY_WIDTH*x+1, KEY_HEIGHT*y+2, 30, 26);
        if (CGRectContainsPoint(rect, point)) {
            NSLog(@"x,y = %d, %d \n", x, y);
            KeyItem item = keyMappingTable[x+y*10];
            if (item.keyCode != INVALID_KEY) {
                NSLog(@"KeyCode: %s\n", item.keyValue);
                keypadmatrix[item.index/8][item.index%8] = 0;
            }
        }
    }    
}

@end
