#ifndef _NEKO_DRIVER_UNIT_H
#define _NEKO_DRIVER_UNIT_H


#include <vector>
#include <string>
#include "Thread.h"
using namespace std;

typedef string QString;
typedef unsigned char QByteArray;

#if FALSE//(TARGET_OS_IPHONE || TARGET_OS_MAC)
typedef void (^OnLCDBufferChanged)(const char* buffer);
#else
typedef void (*OnLCDBufferChanged)(const char* buffer);
#endif


class EmulatorThread : public Thread
{
public:
     EmulatorThread();
    ~EmulatorThread();
protected:
    void* fLCDBuffer;
    OnLCDBufferChanged fLCDBufferChanged;
private:
    unsigned int lastTicket;
    unsigned long long totalcycle;
    //const unsigned spdc1016freq = 3686400;
    bool checked;
    unsigned batchlimiter;
    long batchcount;
    double sleepgap;
    long sleepcount;
    void main();
	
protected:
    virtual void run();

public:
    void SetOnLCDBufferChanged(OnLCDBufferChanged lcdBufferChanged);  
};

class TNekoDriver {
public:
    TNekoDriver();
    ~TNekoDriver();
private:
    string fProjectViewNameFilter;
    EmulatorThread fEmulatorThread;
    char* fNorBuffer; // for performance
    char* fBROMBuffer;
    OnLCDBufferChanged fLCDBufferChanged;
private:
    bool LoadBROM(const QString& filename);
    bool LoadFullNorFlash(const QString& filename);
    bool LoadDemoNor(const QString& filename);
    //static bool NicetoScaleOnFly( QSize &orgsize, bool ispng, bool isjpeg );
    //static bool NiceToRotate( int width, int height, int destwidth, int destheight );
    //static bool PrepareScaledQPicture( QImage& qimage, int destwidth, int destheight, ScaleFilter filter, int& scaletime );
    //static void OptimizeMangaImage( QImage& qimage );
    //static void SaveBundleRec( TPageBundleRec& page ); // Encode to file and drop image
    //bool PreparePageImagesBuffer();
    //void CalulatePages(int& allpagecount, int &readypagecount);
public:
    bool IsProjectEmpty();
    bool IsProjectModified();
    void SwitchNorBank(int bank);
    void Switch4000toBFFF(unsigned char bank); // used 0A/0D value
    void InitInternalAddrs();
    bool StartEmulation();
    bool RunDemoBin(const QString& filename);
    bool StopEmulation();
    bool PauseEmulation();
    bool ResumeEmulation();
	inline bool IsStarted() { return fEmulatorThread.isStarted(); }

    void SetOnLCDBufferChanged(OnLCDBufferChanged lcdBufferChanged);
    void OnStepFinished(unsigned short pc);
};

typedef TNekoDriver* PNekoDriver;
extern PNekoDriver theNekoDriver;

extern unsigned short lcdbuffaddr;

extern unsigned char keypadmatrix[8][8];

#endif