#include "Thread.h"

Thread::Thread():
mThread(NULL),
mStarted(false),
mInterrupted(false)
{
    pthread_cond_init(&mCondition, NULL);
	pthread_mutex_init(&mMutex, NULL);
}

Thread::~Thread()
{
    pthread_cond_destroy(&mCondition);
	pthread_mutex_destroy(&mMutex);
}

void* Thread::startProc(void *param)
{
    Thread *p = (Thread*)param;
    p->run();
	pthread_mutex_lock(&p->mMutex);
	p->mInterrupted = false;
	p->mStarted = false;
	pthread_mutex_unlock(&p->mMutex);
	return NULL;
}

void Thread::start()
{
    pthread_mutex_lock(&mMutex);
    if (!mStarted) {
       int err = pthread_create(&mThread, NULL, startProc, (void*)this);
	   if (err == 0) mStarted = true;
	}
	pthread_mutex_unlock(&mMutex);
}

void Thread::stop()
{
    interrupt();
}

void Thread::join()
{
    pthread_join(mThread, NULL);
}

void Thread::wait()
{
    pthread_mutex_lock(&mMutex);
	pthread_cond_wait(&mCondition, &mMutex);
	pthread_mutex_unlock(&mMutex);
}

void Thread::wait(int msecs)
{
   pthread_mutex_lock(&mMutex);
   pthread_cond_timedwait(&mCondition, &mMutex, NULL);
   pthread_mutex_unlock(&mMutex);
}

void Thread::notify()
{
   pthread_mutex_lock(&mMutex);
   pthread_cond_signal(&mCondition);
   pthread_mutex_unlock(&mMutex);
}

bool Thread::isInterrupted()
{
    return mInterrupted;
}

void Thread::interrupt()
{
    pthread_mutex_lock(&mMutex);
    if (!mInterrupted && mStarted) mInterrupted = true;
	pthread_mutex_unlock(&mMutex);
}

bool Thread::isStarted()
{
    return mStarted;
}